@@ -4,18 +4,12 @@
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 
-import os
 import json
 import struct
-import warnings
 import sys
-
-try:
-    from urllib.parse import urlparse # Py 3
-except ImportError:
-    from urlparse import urlparse # Py 2
-
 import tornado
+
+from urllib.parse import urlparse
 from tornado import gen, ioloop, web
 from tornado.websocket import WebSocketHandler
 
@@ -24,6 +18,8 @@ from jupyter_client.jsonutil import date_default, extract_dates
 from ipython_genutils.py3compat import cast_unicode
 
 from .handlers import JupyterHandler
+from jupyter_server.utils import maybe_future
+
 
 def serialize_binary_message(msg):
     """serialize a message as a binary blob
@@ -280,9 +276,9 @@ class AuthenticatedZMQStreamHandler(ZMQStreamHandler, JupyterHandler):
         # pre_get can be a coroutine in subclasses
         # assign and yield in two step to avoid tornado 3 issues
         res = self.pre_get()
-        yield gen.maybe_future(res)
+        yield maybe_future(res)
         res = super(AuthenticatedZMQStreamHandler, self).get(*args, **kwargs)
-        yield gen.maybe_future(res)
+        yield maybe_future(res)
 
     def initialize(self):
         self.log.debug("Initializing websocket connection %s", self.request.path)
