@@ -1,9 +1,13 @@
 import sys
+import re
+
+from jinja2 import Environment, FileSystemLoader
 
 from traitlets import (
     Unicode, 
     List, 
-    Dict, 
+    Dict,
+    Bool,
     default, 
     validate
 )
@@ -23,8 +27,29 @@ try:
 except KeyError:
     pass
 
+#-----------------------------------------------------------------------------
+# Util functions and classes.
+#-----------------------------------------------------------------------------
 
-def _preparse_command_line(Application):
+def _preparse_for_subcommand(Application, argv):
+    """Preparse command line to look for subcommands.
+    """
+    # Read in arguments from command line.
+    if len(argv) == 0:
+        return
+
+    # Find any subcommands.
+    if Application.subcommands and len(argv) > 0:
+        # we have subcommands, and one may have been specified
+        subc, subargv = argv[0], argv[1:]
+        if re.match(r'^\w(\-?\w)*$', subc) and subc in Application.subcommands:
+            # it's a subcommand, and *not* a flag or class parameter
+            app = Application()
+            app.initialize_subcommand(subc, subargv)
+            return app.subapp
+
+
+def _preparse_for_stopping_flags(Application, argv):
     """Looks for 'help', 'version', and 'generate-config; commands 
     in command line. If found, raises the help and version of 
     current Application.
@@ -38,9 +63,9 @@ def _preparse_command_line(Application):
     # version), we want to only search the arguments up to the first
     # occurrence of '--', which we're calling interpreted_argv.
     try:
-        interpreted_argv = sys.argv[:sys.argv.index('--')]
+        interpreted_argv = argv[:argv.index('--')]
     except ValueError:
-        interpreted_argv = sys.argv
+        interpreted_argv = argv
 
     # Catch any help calls.
     if any(x in interpreted_argv for x in ('-h', '--help-all', '--help')):
@@ -61,6 +86,53 @@ def _preparse_command_line(Application):
         app.exit(0)
 
 
+class ExtensionAppJinjaMixin:
+    """Use Jinja templates for HTML templates on top of an ExtensionApp."""
+    
+    jinja2_options = Dict(
+        help=_("""Options to pass to the jinja2 environment for this
+        extension.
+        """)
+    ).tag(config=True)
+
+    def _prepare_templates(self):
+        # Add templates to web app settings if extension has templates.
+        if len(self.template_paths) > 0:
+            self.settings.update({
+                "{}_template_paths".format(self.extension_name): self.template_paths
+            })
+
+        # Create a jinja environment for logging html templates.
+        self.jinja2_env = Environment(
+            loader=FileSystemLoader(self.template_paths), 
+            extensions=['jinja2.ext.i18n'],
+            autoescape=True,
+            **self.jinja2_options
+        )
+
+        # Get templates defined in a subclass.
+        self.initialize_templates()
+
+        # Add the jinja2 environment for this extension to the tornado settings.
+        self.settings.update(
+            {
+                "{}_jinja2_env".format(self.extension_name): self.jinja2_env 
+            }
+        )
+
+#-----------------------------------------------------------------------------
+# Aliases and Flags
+#-----------------------------------------------------------------------------
+
+flags['no-browser']=(
+    {'ExtensionApp' : {'open_browser' : True}},
+    _("Prevent the opening of the default url in the browser.")
+)
+
+#-----------------------------------------------------------------------------
+# ExtensionApp
+#-----------------------------------------------------------------------------
+
 class ExtensionApp(JupyterApp):
     """Base class for configurable Jupyter Server Extension Applications.
 
@@ -72,24 +144,36 @@ class ExtensionApp(JupyterApp):
         class method. This method can be set as a entry_point in 
         the extensions setup.py
     """
+    # Subclasses should override this trait. Tells the server if
+    # this extension allows other other extensions to be loaded
+    # side-by-side when launched directly.
+    load_other_extensions = True
+
     # Name of the extension
     extension_name = Unicode(
-        "",
         help="Name of extension."
     )
 
-    @default("extension_name")
-    def _default_extension_name(self):
-        raise Exception("The extension must be given a `name`.")
+    def _extension_name_default(self):
+        try:
+            return self.name
+        except AttributeError:
+            raise ValueError("The extension must be given a `name`.")
 
-    @validate("extension_name")
-    def _valid_extension_name(self, obj, value):
-        if isinstance(name, str):
+    INVALID_EXTENSION_NAME_CHARS = [' ', '.', '+', '/']
+
+    @validate('extension_name')
+    def _validate_extension_name(self, value):
+        #value = self.extension_name
+        if isinstance(value, str):
             # Validate that extension_name doesn't contain any invalid characters.
-            for char in [' ', '.', '+', '/']:
-                self.error(obj, value)
+            for c in ExtensionApp.INVALID_EXTENSION_NAME_CHARS:
+                if c in value:
+                    raise ValueError("Extension name '{name}' cannot contain any of the following characters: "
+                                     "{invalid_chars}.".
+                                     format(name=value, invalid_chars=ExtensionApp.INVALID_EXTENSION_NAME_CHARS))
             return value
-        self.error(obj, value) 
+        raise ValueError("Extension name must be a string, found {type}.".format(type=type(value)))
 
     # Extension can configure the ServerApp from the command-line
     classes = [
@@ -99,6 +183,8 @@ class ExtensionApp(JupyterApp):
     aliases = aliases
     flags = flags
 
+    subcommands = {}
+
     @property
     def static_url_prefix(self):
         return "/static/{extension_name}/".format(
@@ -126,10 +212,63 @@ class ExtensionApp(JupyterApp):
         help=_("""Handlers appended to the server.""")
     ).tag(config=True)
 
+    def _config_dir_default(self):
+        """Point the config directory at the server's config_dir by default."""
+        try:
+            return self.serverapp.config_dir
+        except AttributeError:
+            raise AttributeError(
+                "The ExtensionApp has not ServerApp "
+                "initialized. Try `.initialize_server()`."
+            )
+
+    def _config_file_name_default(self):
+        """The default config file name."""
+        if not self.extension_name:
+            return ''
+        return 'jupyter_{}_config'.format(self.extension_name.replace('-','_'))
+
     default_url = Unicode('/', config=True,
         help=_("The default URL to redirect to from `/`")
     )
 
+    open_browser = Bool(
+        True,
+        help=_("Should the extension open a browser window?")
+    )
+
+    custom_display_url = Unicode(u'', config=True,
+        help=_("""Override URL shown to users.
+
+        Replace actual URL, including protocol, address, port and base URL,
+        with the given value when displaying URL to the users. Do not change
+        the actual connection URL. If authentication token is enabled, the
+        token is added to the custom URL automatically.
+
+        This option is intended to be used when the URL to display to the user
+        cannot be determined reliably by the Jupyter server (proxified
+        or containerized setups for example).""")
+    )
+
+    @default('custom_display_url') 
+    def _default_custom_display_url(self):
+        """URL to display to the user."""
+        # Get url from server.
+        url = url_path_join(self.serverapp.base_url, self.default_url)
+        return self.serverapp.get_url(self.serverapp.ip, url)
+
+    def _write_browser_open_file(self, url, fh):
+        """Use to hijacks the server's browser-open file and open at 
+        the extension's homepage.
+        """
+        # Ignore server's url
+        del url
+        path = url_path_join(self.serverapp.base_url, self.default_url)
+        url = self.serverapp.get_url(path=path, token=self.serverapp.token)
+        jinja2_env = self.serverapp.web_app.settings['jinja2_env']
+        template = jinja2_env.get_template('browser-open.html')
+        fh.write(template.render(open_url=url))
+
     def initialize_settings(self):
         """Override this method to add handling of settings."""
         pass
@@ -147,8 +286,8 @@ class ExtensionApp(JupyterApp):
         the object to the webapp's settings as `<extension_name>_config`.  
         """
         traits = self.class_own_traits().keys()
-        self.config = Config({t: getattr(self, t) for t in traits})
-        self.settings['{}_config'.format(self.extension_name)] = self.config
+        self.extension_config = Config({t: getattr(self, t) for t in traits})
+        self.settings['{}_config'.format(self.extension_name)] = self.extension_config
 
     def _prepare_settings(self):
         # Make webapp settings accessible to initialize_settings method
@@ -212,17 +351,16 @@ class ExtensionApp(JupyterApp):
             self.settings.update({
                 "{}_template_paths".format(self.extension_name): self.template_paths
             })
-        self.initialize_templates()
 
     @staticmethod
-    def initialize_server(argv=[], **kwargs):
+    def initialize_server(argv=[], load_other_extensions=True, **kwargs):
         """Get an instance of the Jupyter Server."""
         # Get a jupyter server instance
-        serverapp = ServerApp(**kwargs)
+        serverapp = ServerApp.instance(**kwargs)
         # Initialize ServerApp config.
         # Parses the command line looking for 
         # ServerApp configuration.
-        serverapp.initialize(argv=argv)
+        serverapp.initialize(argv=argv, load_extensions=load_other_extensions)
         return serverapp
 
     def initialize(self, serverapp, argv=[]):
@@ -235,35 +373,40 @@ class ExtensionApp(JupyterApp):
         - Passes settings to webapp
         - Appends handlers to webapp.
         """
+        # Initialize ServerApp.
+        self.serverapp = serverapp
+        
         # Initialize the extension application
         super(ExtensionApp, self).initialize(argv=argv)
-        self.serverapp = serverapp
 
-        # Intialize config, settings, templates, and handlers.
+        # Initialize config, settings, templates, and handlers.
         self._prepare_config()
         self._prepare_templates()
         self._prepare_settings()
         self._prepare_handlers()
 
-    def listen(self):
-        """Extension doesn't listen for anything. IOLoop lives
-        in the attached server.
-        """
-        pass
-
     def start(self):
-        """Extension has nothing to `start`. See `start_server`
-        for starting the jupyter server application.
-        """
-        pass
-
-    def start_server(self, **kwargs):
-        """Start the Jupyter server.
+        """Start the underlying Jupyter server.
         
         Server should be started after extension is initialized.
         """
+        super(ExtensionApp, self).start()
+        # Override the browser open file to 
+        # Override the server's display url to show extension's display URL.
+        self.serverapp.custom_display_url = self.custom_display_url
+        # Override the server's default option and open a broswer window.
+        self.serverapp.open_browser = self.open_browser
+        # Hijack the server's browser-open file to land on
+        # the extensions home page.
+        self.serverapp._write_browser_open_file = self._write_browser_open_file
         # Start the server.
-        self.serverapp.start(**kwargs)
+        self.serverapp.start()
+
+    def stop(self):
+        """Stop the underlying Jupyter server.
+        """
+        self.serverapp.stop()
+        self.serverapp.clear_instance()
 
     @classmethod
     def load_jupyter_server_extension(cls, serverapp, argv=[], **kwargs):
@@ -273,27 +416,6 @@ class ExtensionApp(JupyterApp):
         # Configure and initialize extension.
         extension = cls()
         extension.initialize(serverapp, argv=argv)
-
-        return extension
-
-    @classmethod
-    def _prepare_launch(cls, serverapp, argv=[], **kwargs):
-        """Prepare the extension application for launch by 
-        configuring the server and the extension from argv.
-        Does not start the ioloop.
-        """
-        # Load the extension
-        extension = cls.load_jupyter_server_extension(serverapp, argv=argv, **kwargs)
-        
-        # Start the browser at this extensions default_url, unless user
-        # configures ServerApp.default_url on command line.
-        try:
-            server_config = extension.config['ServerApp']
-            if 'default_url' not in server_config:
-                serverapp.default_url = extension.default_url
-        except KeyError: 
-            pass
-
         return extension
 
     @classmethod
@@ -302,19 +424,32 @@ class ExtensionApp(JupyterApp):
         and appends the extension to the server. Then starts the server and routes to
         extension's landing page.
         """
-        # Check for help, version, and generate-config arguments
-        # before initializing server to make sure these
-        # arguments trigger actions from the extension not the server.
-        _preparse_command_line(cls)
         # Handle arguments.
-        if argv is not None:
+        if argv is None:
             args = sys.argv[1:]  # slice out extension config.
         else:
             args = []
-        
-        # Get a jupyter server instance.
-        serverapp = cls.initialize_server(argv=args)
-        extension = cls._prepare_launch(serverapp, argv=args, **kwargs)
-        # Start the ioloop.
-        extension.start_server()
+        # Check for subcommands
+        subapp = _preparse_for_subcommand(cls, args)
+        if subapp:
+            subapp.start()
+        else:
+            # Check for help, version, and generate-config arguments
+            # before initializing server to make sure these
+            # arguments trigger actions from the extension not the server.
+            _preparse_for_stopping_flags(cls, args) 
+            # Get a jupyter server instance.
+            serverapp = cls.initialize_server(
+                argv=args, 
+                load_other_extensions=cls.load_other_extensions
+            )
+            # Log if extension is blocking other extensions from loading.
+            if not cls.load_other_extensions:
+                serverapp.log.info(
+                    "{ext_name} is running without loading "
+                    "other extensions.".format(ext_name=cls.extension_name)
+                )
 
+            extension = cls.load_jupyter_server_extension(serverapp, argv=args, **kwargs)
+            # Start the ioloop.
+            extension.start()
\ No newline at end of file
